#include <stdio.h>
#include <liba.h>
#include <libb.h>

int
main(void)
{
  printf("start value = %d\n", liba_get());
  liba_add(2);
  libb_mul(5);
  printf("end value = %d\n", liba_get());
  return 0;
}
